blockquote {
    font-style: italic;
}

div.admonition-todo {
    background-color: #ffa;
    border: 1px solid #ee2;
}

div.dbapi-extension {
    background-color: #eef;
    border: 1px solid #aaf;
}

code.sql,
tt.sql {
    font-size: 1em;
    background-color: transparent;
}

a > code.sql,
a > tt.sql {
    font-weight: normal;
}

a > code.sql:hover,
a > tt.sql:hover {
    text-decoration: underline;
}

dl.faq dt {
    font-weight: bold;
}

table.data-types div.line-block {
    margin-bottom: 0;
}


/* better theme customisation */

body {
    background-color: #216464;
}

header, .related, .document, footer {
    background-color: white;
}

header h1 {
    font-size: 150%;
    margin-bottom: 0;
    padding: 0.5rem 10px 0.5rem 10px;
}

h1, h2, h3 {
    font-weight: normal;
}

.body h1, .body h2, .body h3 {
    color: #074848;
}

h1 {
    font-size: 200%;
}

h2 {
    font-size: 160%;
}

h3 {
    font-size: 140%;
}

footer#pagefooter {
    margin-bottom: 1rem;
    font-size: 85%;
    color: #444;
}

#rellinks, #breadcrumbs {
    padding-right: 10px;
    padding-left: 10px;
}

.sphinxsidebar {
    padding-left: 10px;
}

.bodywrapper {
    padding-right: 10px;
}

div.body h1, div.body h2, div.body h3 {
    background-color: #f2f2f2;
    border-bottom: 1px solid #d0d0d0;
}

div.body p.rubric {
    border-bottom: 1px solid #d0d0d0;
}

body .sphinxsidebar .search {
    margin-top: 0;
}

html pre {
    background-color: #efc;
    border: 1px solid #ac9;
    border-left: none;
    border-right: none;
}

a, a:visited {
    color: #0b6868;
}

th {
    background-color: #ede;
}

code.xref, a code {
    font-weight: bold;
}

code.descname {
    font-weight: bold;
    font-size: 120%;
}

@media (max-width: 820px) {
    body {
        background-color: white;
    }
}
